/**********************************************************
	admserver.c
	version 1.1
	Kirill Mechitov (mechitov@cs.uiuc.edu)
**********************************************************/

/* Actor Deployment and Messaging Service */

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <malloc.h>
#include <time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <pthread.h>
#include "parse.h"
#include "thread.h"
#include "hw.h"
#include "admserver.h"
#include <errno.h>

int ssock;

int nthreads = 0;
pthread_mutex_t nthreads_mutex = PTHREAD_MUTEX_INITIALIZER;
pthread_cond_t nthreads_cond = PTHREAD_COND_INITIALIZER;

int nactors = 0;
pthread_mutex_t nactors_mutex = PTHREAD_MUTEX_INITIALIZER;

int asock = -1;
pthread_mutex_t asock_mutex = PTHREAD_MUTEX_INITIALIZER;

pthread_mutex_t msgq_mutex = PTHREAD_MUTEX_INITIALIZER;

typedef struct {
	int key;
	int value;
} map_element;
map_element actor_map[MAX_ACTORS];

typedef struct {
	int value;
} thread_args;

static void errquit(char *msg)
{
	fprintf(stderr, "error: %s\n", msg);
	exit(1);
}

/* thread for reading messages from WSN and notifying meta-actors */
void *message_thread(void *args)
{
	int i;
	int msglen;
	int maid;
	char msgbuf[MAX_MSG_SIZE];

	/* wait for messages */
	for (;;) {
		/* received msg from ActorNet */
		if (read(sp[0], &maid, sizeof (int)) < 0) {
			errquit("read failed - message thread");
		}
		if ((msglen = read(sp[0], msgbuf, MAX_MSG_SIZE)) < 0) {
			printf ("errno = %d\n", errno);
			errquit("read failed - message thread");
		}
		msgbuf[msglen] = 0;

		/* find which meta-actor to signal */
		pthread_mutex_lock(&nactors_mutex);
		for (i = 0; i < nactors; i++) {
			if (maid == actor_map[i].key) { 
				printf("%d: received msg %s\n", maid, msgbuf);
				if (write(actor_map[i].value, msgbuf, msglen+1) < 0) {
					printf ("errno = %d\n", errno);
					errquit("write failed - message thread");
				}
				break;
			}
		}
		pthread_mutex_unlock(&nactors_mutex);
	}

	pthread_exit(NULL);
	return NULL;
}

/* service thread for handling client connections */
void *service_thread(void *args)
{
	int sock = ((thread_args *)args)->value;
	char actor_name[100], maid[20], *ack = "1", eof[3];
	int maid_int, maid_len, neof;
	int i;

	srand(time(NULL) + getpid());

	/* read meta-actor id, send ack */
	if ((maid_len = read(sock, maid, 20)) < 0) {
		errquit("read failed - metaactor id");
	}
	maid[maid_len] = 0;
	maid_int = atoi(maid);
	if (write(sock, ack, strlen(ack) + 1) < 0) {
		errquit("write ack failed - metaactor id");
	}
	printf("%d: received metaactor id\n", maid_int);

	/* register the (meta-actor id, socket) pair */
	pthread_mutex_lock(&nactors_mutex);
	if (nactors < MAX_ACTORS) {
		actor_map[nactors].key = maid_int;
		actor_map[nactors].value = sock;
		nactors++;
	}
	else {
		errquit("too many actors");
	}
	pthread_mutex_unlock(&nactors_mutex);

	/* deploy actor */
	pthread_mutex_lock(&asock_mutex);
	asock = sock;
	printf("%d: executing actor\n", maid_int);
	sprintf(actor_name, "%ld", time(NULL) + rand());
	run(execute, actor_name, 500);
	for (initThread(); resume(); ) { }
	read(sock, eof, 2);
	printf("%d: done executing\n", maid_int);
	pthread_mutex_unlock(&asock_mutex);

	/* wait for meta-actor to unregister by sending EOF */
	if ((neof = read(sock, eof, 2)) < 0) {
		printf ("errno = %d\n", errno);
		errquit("read failed - read eof");
	}
	eof[2] = 0;
    if (neof < 2 || strcmp(eof, "-1") != 0) {
		errquit("unexpected input");
	}
	pthread_mutex_lock(&nactors_mutex);
	for (i = 0; i < nactors; i++) {
		if (actor_map[i].key == maid_int) {
			actor_map[i].key = actor_map[nactors - 1].key;
			actor_map[i].value = actor_map[nactors - 1].value;
			nactors--;
			break;
		}
	}
	pthread_mutex_unlock(&nactors_mutex);

	/* close connection */
	printf("%d: closing connection\n", maid_int);
	close(sock);
	free(args);
	pthread_mutex_lock(&nthreads_mutex);
	nthreads--;
	if (nthreads == MAX_THREADS - 1) {
		pthread_cond_signal(&nthreads_cond);
	}
	pthread_mutex_unlock(&nthreads_mutex);
	pthread_exit(NULL);
	return NULL;
}

void admserver_start(int port)
{
	struct sockaddr_in sa;
	pthread_t thread;
	pthread_attr_t attr;

	pthread_attr_init(&attr);
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);

	/* spawn message reader thread */
	if (pthread_create(&thread, &attr, message_thread, NULL) != 0) {
		errquit("pthread_create failed");
	}

	/* initialize server address struct */
	memset(&sa, 0, sizeof (struct sockaddr_in));
	sa.sin_family = AF_INET;
	sa.sin_addr.s_addr = htonl(INADDR_ANY);
	sa.sin_port = htons(port);

	/* create server socket and listen for connections */
	if ((ssock = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
		errquit("socket failed");
	}
	if (bind(ssock, (struct sockaddr*) &sa, sizeof(sa)) < 0) {
		errquit("bind failed");
	}
	if (listen(ssock, BACKLOG) < 0) {
		errquit("listen failed");
	}

	for (;;) {
		int sock;
		struct sockaddr_in ca;
		socklen_t len;
		thread_args *ta;

		/* check for available threads */
		pthread_mutex_lock(&nthreads_mutex);
		while (nthreads >= MAX_THREADS
			&& pthread_cond_wait(&nthreads_cond, &nthreads_mutex) != 0);
		nthreads++;
		pthread_mutex_unlock(&nthreads_mutex);

		/* get client connection */
		len = sizeof (struct sockaddr);
		if ((sock = accept(ssock, (struct sockaddr*) &ca, &len)) < 0) {
			errquit("accept failed");
		}
		printf("<--- connection: %d threads left\n", MAX_THREADS - nthreads);

		/* spawn service thread */
		if ((ta = malloc(sizeof (thread_args))) == NULL) {
			errquit("malloc failed");
		}
		ta->value = sock;
		if (pthread_create(&thread, &attr, service_thread, ta) != 0) {
			pthread_mutex_lock(&nthreads_mutex);
			nthreads--;
			pthread_mutex_unlock(&nthreads_mutex);
		}
	}
}
