/**********************************************************
	pc.c
	version 1.1
	YoungMin Kwon (ykwon4@cs.uiuc.edu)
	Kirill Mechitov (mechitov@cs.uiuc.edu)
**********************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <termios.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/select.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <pthread.h>
#include <errno.h>
#include "memory.h"
#include "data.h"
#include "thread.h"
#include "comm.h"
#include "eval.h"
#include "hw.h"
#include "parse.h"
#include "packet.h"
#include "prim.h"
#if defined(PLATFORM) || defined(ADMSERVER)
#include "init.h"
#endif
#if defined(ADMSERVER)
#include "admserver.h"

int sp[2];
pid_t pid;
#endif



static uint8_t flash[FlashSize];
static uint8_t recvDone;
int rd, wd, hwid;

static const uint16_t *beamdata;
static const uint16_t beamdata200[20] = {
									0, 0, 0, 0, 1, 1, 1, 0, 0, 1,
									1, 1, 0, 0, 0, 0, 0, 0, 0, 0
									};
static const uint16_t beamdata300[20] = {
									0, 0, 0, 0, 0, 1, 1, 1, 0, 0,
									0, 1, 1, 1, 0, 0, 0, 0, 0, 0
									};
static const uint16_t beamdata500[20] = {
									0, 0, 0, 0, 0, 0, 1, 1, 1, 0,
									0, 0, 0, 1, 1, 1, 0, 0, 0, 0
									};
static const uint16_t photodata[20] = {
									10, 10, 10, 10, 10, 10, 10, 10, 10, 10,
									11, 11, 11, 11, 11, 11, 11, 11, 11, 11
									};
static const uint16_t tempdata[20] = {
									0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
									1, 1, 1, 1, 1, 1, 1, 1, 1, 1
									};

static void errquit(char *msg)
{
	fprintf(stderr,"error: %s\n", msg);
	exit(1);
}

uint16_t hwId()
{
	return hwid;
}

uint8_t isSendPending()
{
	return 0;
}

void sendPacket(uint8_t *buf, uint8_t n)
{
  //debugPrint("sendPacket", buf, n);
	Msg msg;
	int i, j;
	uint16_t adrs=0xffff;
	msg.adrs[0]=adrs & 0xff;
	msg.adrs[1]=(adrs >> 8) & 0xff;
	msg.type=10;
	msg.group=0x7d;
	msg.length=n;
	memcpy(msg.data, buf, n);
	for(i=0; i<PacketHeaderLength+n; i+=j)
		j=write(wd, ((uint8_t*)&msg)+i, PacketHeaderLength+n-i);
}

uint8_t isRecvPending()
{
	fd_set rfds;
	struct timeval tv;
	FD_ZERO(&rfds);
	FD_SET(rd, &rfds);
	//tv.tv_sec=0; tv.tv_usec=10;
	tv.tv_sec=0; tv.tv_usec=0;
	return select(rd+1, &rfds, NULL, NULL, &tv);
}

void recvPacket()
{
	Msg msg;
	uint8_t i, *p;
	while(!isRecvPending())
		yield();
	p=(uint8_t*)&msg;
	for(i=0; i<PacketHeaderLength; ) 				//header
		i+=read(rd, p+i, PacketHeaderLength-i);
	p=msg.data;
	for(i=0; i<msg.length; )						//data without crc
		i+=read(rd, p+i, msg.length-i);
	addRawBuf(msg.data, msg.length);
}

void *recvThread(void* arg)
{
	for(recvDone=0; !recvDone; ) {
		if(isRecvPending() && !isRawBufFull())
			recvPacket();
#if defined(__CYGWIN__)
		sched_yield();
#else
		pthread_yield();
#endif
	}
	return arg;
}

uint16_t IO(uint16_t a, uint16_t b)
{
	static uint16_t tempidx = 0;
	static uint16_t photoidx = 0;
	static uint16_t beamidx = 0;
	uint16_t tmp;
	yield();
	switch(a) {
		case 0:	return hwId();
		case 1: tmp = tempdata[tempidx];	// fake temp
				tempidx = (tempidx + 1) % 20;
				return tmp;
		case 2: tmp = photodata[photoidx];	// fake photo
				photoidx = (photoidx + 1) % 20;
				return tmp;
		case 3: {
			struct timeval tv;
			gettimeofday(&tv, NULL);
			return (tv.tv_sec*10+tv.tv_usec/1000/100)&0xffff;
		}
		case 4: tmp = beamdata[beamidx];	// fake beam
				beamidx = (beamidx + 1) % 20;
				return tmp;
		case 9: {	// for debug purposes
			printlist(b);
			return b;
		}
	}
	return 0;
}

void startBeeping()
{
	isBeeping = 1;
}

void stopBeeping()
{
	isBeeping = 0;
}

int _printmsg(uint16_t args, char *msgbuf, int *msgidx)
{
	if (args != nilAdrs) {
		char tmp[20];
		sprintf(tmp, "%d", (*((uint16_t *)deref(car(args)))));
		sprintf(msgbuf + *msgidx, "%s ", tmp);
		*msgidx = *msgidx + strlen(tmp) + 1;
		_printmsg(cdr(args), msgbuf, msgidx);
	}
	return 0;
}

void printmsg(uint16_t args)
{
#ifdef ADMSERVER
	char msgbuf[MAX_MSG_SIZE];
	int msgidx;

	if (args != nilAdrs) {
		int maid;
		maid = *((uint16_t *)deref(car(args)));
		msgidx = 0;
		if (cadr(args) == nilAdrs) {
			// error
			sprintf(msgbuf, "-1");
			msgidx += 2;
		}
		else {
			_printmsg(cadr(args), msgbuf, &msgidx);
		}
		msgbuf[msgidx] = 0;

		pthread_mutex_lock(&msgq_mutex);
		printf("%d: sending msg %s\n", maid, msgbuf);
		if (write(sp[1], &maid, sizeof (int)) < 0) {
			errquit("write failed - 123");
		}
		if (write(sp[1], msgbuf, msgidx+1) < 0) {
			printf("errno: %d\n", errno);
			errquit("write failed - 456");
		}
		pthread_mutex_unlock(&msgq_mutex);
	}
#else
	char msgbuf[10000];
	int msgidx;

	if (args != nilAdrs) {
		int maid;
		maid = *((uint16_t *)deref(car(args)));
		msgidx = 0;
		if (cadr(args) == nilAdrs) {
			// error
			sprintf(msgbuf, "-1");
			msgidx += 2;
		}
		else {
			_printmsg(cadr(args), msgbuf, &msgidx);
		}
		msgbuf[msgidx] = 0;
		fprintf(stderr, "extmsg %d |%s|\n", maid, msgbuf);
	}
#endif
}

void ledSet(uint16_t n)
{
	fprintf(stderr, "---> %d\n", n);
}

void ledToggle(uint16_t n)
{
//	fprintf(stderr,"~%d\n", n);
}

void flashWrite(uint16_t addr, void *p, uint8_t n)
{
	memcpy(&flash[addr], p, n);
	yield();
}

void flashRead(uint16_t addr, void *p, uint8_t n)
{
	memcpy(p, &flash[addr], n);
	yield();
}

static int keywordIndex(uint16_t a)
{
	uint16_t *p=&sym.nil;
	int i;
	for(i=0; i<sizeof(Symbols)/sizeof(uint16_t); i++)
		if(p[i]==a) 
			return i;
	return -1;
}

static void _mark(uint16_t a)
{
	uint8_t *p=fetch(a,1);
	if(a==baseEnv) {
		fprintf(stderr,"baseEnv.%d ", a);
		return;
	}
	switch(p[0] & 0x2f) {
		case TagSymbol: {
			int i=keywordIndex(a);
			i<0?  fprintf(stderr,"s.%d ", a): fprintf(stderr,"%s.%d ", keyword[i], a);
			p[0]^=0x20; break;
		}
		case TagByte:   fprintf(stderr,"b.%d ", a); p[0]^=0x20; break;
		case TagWord:   fprintf(stderr,"%d.%d ", *(uint16_t*)(p+1), a); p[0]^=0x20; break;
		case TagDWord:
		case TagFloat:
		case TagBytes: fprintf(stderr,"B.%d ", a); p[0]^=0x20; break;
		case TagPair: {
			Pair q=*(Pair*)p;
			p[0]^=0x20;
			fprintf(stderr,"( ");
			_mark(q.first);
			fprintf(stderr,", ");
			_mark(q.second);
			fprintf(stderr," ).%d", a);
			break;
		}
		default:
			if((p[0]&0x0f)==TagSymbol) {
				int i=keywordIndex(a);
				i<0?  fprintf(stderr,"!s.%d ", a): fprintf(stderr,"!%s.%d ", keyword[i], a);
			}
			else if(p[0]&0x20) 
				fprintf(stderr,"!.%d", a);
			else
				fprintf(stderr,"?.%d ", a);
	}
}

static void _unmark(uint16_t a)
{
	uint8_t *p=fetch(a,1);
	if(!(p[0] & 0x20))
		return;
	switch(p[0] & 0x0f) {
		case TagSymbol: 
		case TagByte:   
		case TagWord:   
		case TagDWord:
		case TagFloat:
		case TagBytes:	p[0] ^= 0x20; break;
		case TagPair: {
			Pair q=*(Pair*)p;
			p[0]^=0x20;
			_unmark(q.first);
			_unmark(q.second);
			break;
#ifdef DEBUG
		default:
			fprintf(stderr,"error. printlist: unknown tag %d.\n", p[0]^0x20);
#endif
			
		}
	}
}

uint16_t printlist(uint16_t a)
{
	_mark(a);
	_unmark(a);
	return a;
}

int openSoc(char *server, int port)
{
	int sd;
	struct sockaddr_in sa;
	struct hostent *host;

	if((host=gethostbyname(server))==NULL)          //internet
		errquit("gethostbyname");
	memset(&sa, 0, sizeof(sa));
	sa.sin_family=AF_INET;
	memcpy(&sa.sin_addr, host->h_addr, host->h_length);
	sa.sin_port=htons(port);
	if((sd=socket(AF_INET, SOCK_STREAM, 0))<0)
		errquit("socket");
	if(connect(sd, (struct sockaddr*) &sa, sizeof(sa))<0)
		errquit("connect");
	return sd;
}

int main(int argc, char **argv)
{
	pthread_t thread;

#if defined(PLATFORM)
	if (argc != 4) {
		fprintf(stderr, "usage: %s <hw id> <fwd addr> <fwd port>\n", argv[0]);
		exit(0);
	}
#elif defined(ADMSERVER)
	if (argc != 5) {
		fprintf(stderr, "usage: %s <hw id> <fwd addr> <fwd port> <port>\n", argv[0]);
		exit(0);
	}
#else
	if (argc != 5) {
		fprintf(stderr, "usage: %s <hw id> <fwd addr> <fwd port> <actor>\n", argv[0]);
		exit(0);
	}
#endif

	hwid = atoi(argv[1]);
	switch (hwid) {
	case 200:
		beamdata = beamdata200;
		break;
	case 300:
		beamdata = beamdata300;
		break;
	case 500:
		beamdata = beamdata500;
		break;
	default:
		beamdata = beamdata200;
	}

#if defined(ADMSERVER)
	if (socketpair(AF_INET, SOCK_STREAM, 0, sp) < 0) {
		errquit("socketpair failed");
	}
	if ((pid = fork()) < 0) {
		errquit("fork failed");
	}

	/* parent */
	if (pid != 0) {
		close(sp[1]);
		wd = rd = openSoc(argv[2], atoi(argv[3]));
		admserver_start(atoi(argv[4]));
		for (initThread(); resume(); ) { }
		close(rd);
		close(sp[0]);
		return 0;
	}

	/* child: start platform */
	close(sp[0]);
#endif

	wd = rd = openSoc(argv[2], atoi(argv[3]));
	pthread_create(&thread, NULL, recvThread, NULL);

#if defined(PLATFORM) || defined(ADMSERVER)
	run(init, "PLATFORM", 500);
#else
	run(execute, argv[4], 500);
#endif

	for (initThread(); resume(); ) { }
	close(rd);
#if defined(ADMSERVER)
	close(sp[1]);
#endif
	return 0;
}
