/**********************************************************
	memory.c
	version 1.1
	YoungMin Kwon (ykwon4@cs.uiuc.edu)
**********************************************************/
#ifdef PC
#include <stdio.h>
#endif
#include <string.h>
#include "memory.h"
#include "thread.h"
#include "hw.h"

static uint8_t ram[NumPage][sizeof(Page)];			//1 KB
static InversePageTable iptbl[NumPage+1];			//8+head
uint16_t reserveStartPage, reserveEndPage;			//from where search begins and ends for reservation

//virtual memory
void initVM() {
	uint16_t i;
	uint32_t a;
	Page *p;
	memset(ram,   0, NumPage*sizeof(Page));
	memset(iptbl, 0, (NumPage+1)*sizeof(InversePageTable));
	for(a=0; a<FlashSize; a+=sizeof(Page)) 		//initialize flash memory
		flashWrite(a, ram[0], sizeof(Page));
	for(i=0; i<NumPage+1; i++) {
		iptbl[i].next=i+1;						//last page table indicates HeadPage
		iptbl[i].page=i;
	}
	iptbl[HeadPage].next=0;						//head page indicates first page
	p=(Page*)ram[0];
	p->bits.flag.dirty=1;						//keep null pointer reserved
	p->bitmap[0]=1;
	reserveStartPage=reserveEndPage=0;
}
static inline uint16_t _offset(uint16_t addr) {
	return addr % PageSize;
}
static inline uint16_t _page(uint16_t addr) {
	return addr / PageSize;
}
uint8_t* fetch(uint16_t addr, uint8_t wr) {
	uint16_t page=_page(addr);
	uint8_t i, j, u=0, v=0; //i: current page, j: previous page, u: last unlocked page
	Page *pg;
	for(i=iptbl[HeadPage].next, j=HeadPage; 1; j=i, i=iptbl[i].next) {
		pg=(Page*)ram[i];
		if(pg->bits.flag.lock==0)
			u=i, v=j;
		if(iptbl[i].page==page) {		//hit
#ifdef STAT
			stat.hit++;
#endif
			break;
		}
		if(iptbl[i].next==HeadPage) {	//miss
#ifdef STAT
			stat.miss++;
#endif
			i=u, j=v;						//load at the last unlocked page
			pg=(Page*)ram[i];
			if(pg->bits.flag.dirty) {
#ifdef STAT
				stat.wr++;
#endif
				pg->bits.flag.dirty=0;
				flashWrite(iptbl[i].page*sizeof(Page), ram[i], sizeof(Page));
			}
#ifdef STAT
			stat.rd++;
#endif
			flashRead(page*sizeof(Page), ram[i], sizeof(Page));
			iptbl[i].page=page;
			break;
		}
	}
	if(j!=HeadPage) {	//j is already the most recently fetched page
		iptbl[j].next=iptbl[i].next;
		iptbl[i].next=iptbl[HeadPage].next;
		iptbl[HeadPage].next=i;
	}
	if(wr)
		((Page*)ram[i])->bits.flag.dirty=1;
	return &ram[i][_offset(addr)];
}

//dynamic allocation
static uint16_t _mark(uint16_t addr, uint8_t n, uint8_t set) {
	Page *page;
	uint16_t o, i, j, c;
	fetch(addr, 1);
	page=(Page*)ram[iptbl[HeadPage].next];
	o=_offset(addr), i=o/8, j=o%8, c=0;
	while(c<n) {
		if(set) page->bitmap[i] |= (1<<j);
		else	page->bitmap[i] &= (0xff^(1<<j));
		if(++j>=8)
			j=0, i++;
		c++;
	}
	return addr;
}

static uint8_t countFree(Page *page, int n) { //look for a free space of size n in page
	uint8_t c, j, k;
	for(c=0, j=0; j<14; j++) {	//count the number of consecutive free bytes
		for(k=0; k<8; k++) {
			if((page->bitmap[j]&(1<<k))==0) {
				c++;
				if(c>=n) 
					return j*8+k+1-n;
			}
			else
				c=0;
		}
	}
	return 0xff;
}
uint16_t reserve(uint8_t n) {		//n should be less than 112
	uint16_t p=reserveStartPage;
	uint16_t i, j, u=0;
	Page *page;
	if(n<= 0 || n>PageSize) {		//112 + 14 + 2 = 128 (mem + map)
#ifdef DEBUG
		fprintf(stderr, "error. reserve: illigal reservation size %d\n", n);
#endif
		return 0;
	}
	for(i=iptbl[HeadPage].next; i!=HeadPage; i=iptbl[i].next) {
		page=(Page*)ram[i];
		if(page->bits.flag.lock==0)
			u=i;
		j=countFree(page, n);
		if(j!=0xff) {
			reserveStartPage=iptbl[i].page;
			return _mark((iptbl[i].page*PageSize)+j, n, 1);
		}
	}
	i=u;							//load at the last unlocked page
	page=(Page*)ram[i];
	if(page->bits.flag.dirty) {
#ifdef STAT
		stat.wr++;
#endif
		page->bits.flag.dirty=0;
		flashWrite(iptbl[i].page*sizeof(Page), ram[i], sizeof(Page));
	}
	while(1) {
		do {
			p=(p+1)%(FlashSize/sizeof(Page));
			if(p==reserveEndPage) 
				return 0;
			for(j=iptbl[HeadPage].next; j!=HeadPage; j=iptbl[j].next)	//if not already loaded
				if(iptbl[j].page==p)
					break;
		} while(iptbl[j].page==p);
#ifdef STAT
		stat.rd++;
#endif
		flashRead(p*sizeof(Page), page, sizeof(Page));
		iptbl[i].page=p;			//dont' touch iptbl[i].next
		j=countFree(page, n);
		if(j!=0xff) {
			reserveStartPage=iptbl[i].page;
			return _mark((iptbl[i].page*PageSize)+j, n, 1);
		}
	}
}
void release(uint16_t addr, uint8_t n) {
	_mark(addr, n, 0);
}
uint8_t isReserved(uint16_t addr) {
	Page *page;
	uint16_t o=_offset(addr);
	fetch(addr, 0);
	page=(Page*)ram[iptbl[HeadPage].next];
	return page->bitmap[o/8] & (1<<(o%8));
}
uint8_t *lock(uint16_t addr, uint8_t rw) {
	Page *page;
	uint8_t *p=fetch(addr, rw);
	page=(Page*)ram[iptbl[HeadPage].next];
	page->bits.flag.lock++;
	return p;
}
uint8_t *unlock(uint16_t addr, uint8_t rw) {
	Page *page;
	uint8_t *p=fetch(addr, rw);
	page=(Page*)ram[iptbl[HeadPage].next];
	if(page->bits.flag.lock>0)
		page->bits.flag.lock--;
#ifdef DEBUG
	else
		fprintf(stderr,"error. unlock: negative lock count\n");
#endif
	return p;
}
/*
int main(int argc, char **argv) {
	uint16_t i, j, p, q;
	initVM();

	p=reserve(100);
	for(i=0; i<20; i+=2) {
		*((uint16_t*)fetch(p+i, 1))=q=reserve(100);
		for(j=0; j<20; j+=2) {
			*((uint16_t*)fetch(*((uint16_t*)fetch(p+i, 1))+j, 1))=i*j;
		}
	}
	for(i=0; i<20; i+=2) {
		for(j=0; j<20; j+=2) {
			printf("%d:%d\n",i*j,*((uint16_t*)fetch(*((uint16_t*)fetch(p+i, 1))+j, 1)));
		}
	}
}
*/

