/**********************************************************
	prim.c
	version 1.1
	Kirill Mechitov (mechitov@cs.uiuc.edu)
**********************************************************/
#ifdef PC
#include <stdio.h>
#endif
#include <stdarg.h>
#include <math.h>
#include "memory.h"
#include "data.h"
#include "eval.h"
#include "comm.h"
#include "thread.h"
#include "hw.h"
#include "prim.h"

uint8_t isBeeping;

typedef struct {
	uint16_t type;
	uint16_t sensor;
	uint16_t start;
	uint16_t duration;
} Pulse;

typedef struct {
	uint16_t type;
	Pulse p1;
	Pulse p2;
} PulsePair;

typedef struct {
	uint16_t x, y, z;
} Location;

typedef struct {
	uint16_t type;
	uint16_t length;
	uint16_t speed;
} MobileObject;

typedef struct {
	uint16_t type;
	uint16_t length;
	uint16_t speed;
} Vehicle;


uint16_t PRIM(uint16_t a, uint16_t b)
{
	yield();
	switch(a) {
		case 0:	return 0;
		case 1:	return detectBeamEvent();
		case 2:	return sortEdges(b);
		case 3:	return averagePulsePairs(b);
		case 4:	return estimateMotion(b);
		case 5:	return classifyVehicle(b);
		case 6: return photoDetect();
		case 7: return detectShadow();
		case 8: return beep();
		default: {
#ifdef PC
			fprintf(stderr, "unknown primitive call\n");
#endif
			return 0;
		}
	}
	return 0;
}


Location getLocation(uint16_t sensor)
{
	Location l;
	switch (sensor) {
	case 200:
		l.x = 10;
		l.y = 0;
		l.z = 0;
		break;
	case 250:	// avg 200, 300
		l.x = 15;
		l.y = 0;
		l.z = 0;
		break;
	case 300:
		l.x = 20;
		l.y = 0;
		l.z = 0;
		break;
	case 350:	// avg 200, 500
		l.x = 20;
		l.y = 0;
		l.z = 0;
		break;
	case 400:	// avg 300, 500
		l.x = 25;
		l.y = 0;
		l.z = 0;
		break;
	case 500:
		l.x = 30;
		l.y = 0;
		l.z = 0;
		break;
	case 100:	// fallthrough
	default:
		l.x = l.y = l.z = 0;
	}
	return l;
}

uint16_t detectBeamEvent()
{
	static uint16_t idx = 0;
	uint16_t start, duration;

	while (IO(4, 0) == 0) {
		idx++;
	}
	start = ++idx;

	while (IO(4, 0) == 1) {
		idx++;
	}
	duration = ++idx - start;

#ifdef PC
	fprintf(stderr, "detectBeamEvent: (%d %d %d %d)\n", 11, IO(0, 0),
		start, duration);
#endif
	return list(4, newWord(11), newWord(IO(0,0)), newWord(start),
		newWord(duration));
}

uint16_t sortEdges(uint16_t args)
{
	uint16_t a1, a2;
	Pulse p1, p2, *first, *second;

	if (args == nilAdrs || car(args) == nilAdrs || car(car(args)) == nilAdrs
	|| cadr(car(args)) == nilAdrs) {
#ifdef PC
	fprintf(stderr, "sortEdges: nil\n");
#endif
		return nilAdrs;
	}

	a1 = car(car(args));
	p1.type = *(uint16_t *)deref(car(a1));
	a2 = cadr(car(args));
	p2.type = *(uint16_t *)deref(car(a2));
	if (p1.type != 11 || p2.type != 11) {
#ifdef PC
	fprintf(stderr, "sortEdges: nil\n");
#endif
		return nilAdrs;
	}

	p1.sensor = *(uint16_t *)deref(cadr(a1));
	p2.sensor = *(uint16_t *)deref(cadr(a2));
	if (p1.sensor == p2.sensor) {
#ifdef PC
	fprintf(stderr, "sortEdges: nil\n");
#endif
		return nilAdrs;
	}

	p1.start = *(uint16_t *)deref(caddr(a1));
	p1.duration = *(uint16_t *)deref(cadddr(a1));
	p2.start = *(uint16_t *)deref(caddr(a2));
	p2.duration = *(uint16_t *)deref(cadddr(a2));
	if (p1.start < p2.start) {
		first = &p1;
		second = &p2;
	}
	else {
		first = &p2;
		second = &p1;
	}

#ifdef PC
	fprintf(stderr, "sortEdges: (%d %d %d %d %d %d %d)\n", 12, first->sensor,
		first->start, first->duration, second->sensor, second->start,
		second->duration);
#endif
	return list(7, newWord(12), newWord(first->sensor), newWord(first->start),
		newWord(first->duration), newWord(second->sensor),
		newWord(second->start), newWord(second->duration));
}

uint16_t averagePulsePairs(uint16_t args)
{
	uint16_t a1, a2;
	PulsePair pp1, pp2;

	if (args == nilAdrs || car(args) == nilAdrs || car(car(args)) == nilAdrs
	|| cadr(car(args)) == nilAdrs) {
#ifdef PC
	fprintf(stderr, "averagePulsePairs: nil\n");
#endif
		return nilAdrs;
	}

	a1 = car(car(args));
	pp1.type = *(uint16_t *)deref(car(a1));
	a2 = cadr(car(args));
	pp2.type = *(uint16_t *)deref(car(a2));
	if (pp1.type != 12 || pp2.type != 12) {
#ifdef PC
	fprintf(stderr, "averagePulsePairs: nil\n");
#endif
		return nilAdrs;
	}

	pp1.p1.sensor = *(uint16_t *)deref(cadr(a1));
	pp1.p1.start = *(uint16_t *)deref(caddr(a1));
	pp1.p1.duration = *(uint16_t *)deref(cadddr(a1));
	pp1.p2.sensor = *(uint16_t *)deref(cadddr(cdr(a1)));
	pp1.p2.start = *(uint16_t *)deref(cadddr(cddr(a1)));
	pp1.p2.duration = *(uint16_t *)deref(cadddr(cdddr(a1)));

	pp2.p1.sensor = *(uint16_t *)deref(cadr(a2));
	pp2.p1.start = *(uint16_t *)deref(caddr(a2));
	pp2.p1.duration = *(uint16_t *)deref(cadddr(a2));
	pp2.p2.sensor = *(uint16_t *)deref(cadddr(cdr(a2)));
	pp2.p2.start = *(uint16_t *)deref(cadddr(cddr(a2)));
	pp2.p2.duration = *(uint16_t *)deref(cadddr(cdddr(a2)));

	pp1.p1.sensor = (pp1.p1.sensor + pp2.p1.sensor) / 2;
	pp1.p1.start = (pp1.p1.start + pp2.p1.start) / 2;
	pp1.p1.duration = (pp1.p1.duration + pp2.p1.duration) / 2;
	pp1.p2.sensor = (pp1.p2.sensor + pp2.p2.sensor) / 2;
	pp1.p2.start = (pp1.p2.start + pp2.p2.start) / 2;
	pp1.p2.duration = (pp1.p2.duration + pp2.p2.duration) / 2;

	pp1.p1.duration = pp1.p2.duration = (pp1.p1.duration + pp1.p2.duration) / 2;

#ifdef PC
	fprintf(stderr, "averagePulsePairs: (%d %d %d %d %d %d %d)\n", 12,
		pp1.p1.sensor, pp1.p1.start, pp1.p1.duration, pp1.p2.sensor,
		pp1.p2.start, pp1.p2.duration);
#endif
	return list(7, newWord(12), newWord(pp1.p1.sensor), newWord(pp1.p1.start),
		newWord(pp1.p1.duration), newWord(pp1.p2.sensor),
		newWord(pp1.p2.start), newWord(pp1.p2.duration));
}

uint16_t estimateMotion(uint16_t args)
{
	uint16_t a;
	PulsePair app;
	MobileObject mo;
	Location l1, l2;
	uint16_t distance, time;

	if (args == nilAdrs || car(args) == nilAdrs || car(car(args)) == nilAdrs) {
#ifdef PC
	fprintf(stderr, "estimateMotion: nil\n");
#endif
		return nilAdrs;
	}

	a = car(car(args));
	app.type = *(uint16_t *)deref(car(a));
	if (app.type != 12) {
#ifdef PC
	fprintf(stderr, "estimateMotion: nil\n");
#endif
		return nilAdrs;
	}

	app.p1.sensor = *(uint16_t *)deref(cadr(a)); 
	app.p1.start = *(uint16_t *)deref(caddr(a)); 
	app.p1.duration = *(uint16_t *)deref(cadddr(a)); 
	l1 = getLocation(app.p1.sensor);

	app.p2.sensor = *(uint16_t *)deref(cadddr(cdr(a))); 
	app.p2.start = *(uint16_t *)deref(cadddr(cddr(a))); 
	app.p2.duration = *(uint16_t *)deref(cadddr(cdddr(a))); 
	l2 = getLocation(app.p2.sensor);

	distance = sqrt((l2.x - l1.x)*(l2.x - l1.x) + (l2.y - l1.y)*(l2.y - l1.y)
		+ (l2.z - l1.z)*(l2.z - l1.z));
	time = app.p2.start - app.p1.start;
	mo.speed = distance / time;
	mo.length = app.p1.duration * mo.speed;

#ifdef PC
	fprintf(stderr, "estimateMotion: (%d %d %d)\n", 13, mo.length, mo.speed);
#endif
	return list(3, newWord(13), newWord(mo.length), newWord(mo.speed));
}

uint16_t classifyVehicle(uint16_t args)
{
	uint16_t a;
	MobileObject mo;

	if (args == nilAdrs || car(args) == nilAdrs || car(car(args)) == nilAdrs) {
#ifdef PC
	fprintf(stderr, "classifyVehicle nil\n");
#endif
		return nilAdrs;
	}

	a = car(car(args));
	mo.type = *(uint16_t *)deref(car(a));
	if (mo.type != 13) {
#ifdef PC
	fprintf(stderr, "classifyVehicle: nil\n");
#endif
		return nilAdrs;
	}

	mo.length = *(uint16_t *)deref(cadr(a)); 
	mo.speed = *(uint16_t *)deref(caddr(a)); 

	/* FIXME - for debugging query engine
	if (mo.speed < 7 || mo.length < 20) {
#ifdef PC
	fprintf(stderr, "classifyVehicle: nil\n");
#endif
		return nilAdrs;
	}
	*/

#ifdef PC
	fprintf(stderr, "classifyVehicle: (%d %d %d)\n", 14, mo.length, mo.speed);
#endif
	return list(3, newWord(14), newWord(mo.length), newWord(mo.speed));
}

uint16_t photoDetect()
{
	uint16_t data;
	uint8_t leds;

	for (;;) {
		leds = 0;
		data = IO(2, 0);
		if (data < 880) {
			leds |= 1;
		}
		if (data < 900) {
			leds |= 2;
		}
		if (data < 920) {
			leds |= 4;
		}
		ledSet(leds);
	}
#ifdef PC
	fprintf(stderr, "photoDetect: (%d %d)\n", 23, data);
#endif

	return list(2, newWord(23), newWord(data));
}

uint16_t detectShadow()
{
	int i;
	uint16_t data;
	uint16_t avg = 0;
	ledSet(4);
	for (i = 0; i < 10; i++) {
		avg += IO(2, 0);
	}
	avg /= 10;
	while ((data = IO(2, 0)) > avg - 15) {
		yield();
	}
	ledSet(7);
#ifdef PC
	fprintf(stderr, "detectShadow: (%d %d)\n", 20, data);
#endif
	return list(2, newWord(20), newWord(data));
}

uint16_t beep()
{
	startBeeping();
#ifdef PC
	fprintf(stderr, "beep: (%d %d)\n", 21, 100);
#endif
	return list(2, newWord(21), newWord(5));
}
